/*
* Usman Dastgeer, PELAB, IDA
* Multicore Computing Course, Sep 2013.
* A simple Map skeleton example for an integer array (can be made generic using e.g. C++ templates) in CUDA.
* The code is just for demonstration and is not well-written (e.g. no proper usage of const keyword etc.)
*/

#include <cuda.h>
#include <iostream>

// Helpers for checking errors from CUDA calls 
#define checkCudaErrors(val) check( (val), #val, __FILE__, __LINE__)

template<typename T>
void check(T err, const char* const func, const char* const file, const int line) {
   if (err != cudaSuccess) {
      std::cerr << "CUDA error at: " << file << ":" << line << std::endl;
      std::cerr << cudaGetErrorString(err) << " " << func << std::endl;
      exit(1);
   }
}


// user function
struct MapFunc
{
   // CUDA device function
   __device__ int cuda_func(int a1, int a2)
   {
      return a1+a2;
   } 
};

// CUDA Kernel
__global__ void MapKernel_CU(MapFunc mapFunc, int* input1, int*input2, int* output, unsigned int n)
{
   unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;

   if(i < n)
   output[i] = mapFunc.cuda_func(input1[i], input2[i]);
}

// skeleton function
void map_func(int *input1, int *input2, int *output, const int size, MapFunc mfunc)
{
   // it's not portable to hard code threadsize as different machines may have different limit for maximum threadsize that you can have
   // which you can know by quering device properties but here we keep it simple.
   int numThreads = 256;
   int numBlocks = (size+numThreads-1)/numThreads;
   size_t sizeVec = size*sizeof(int);
   
   // memory allocation for dev_in1 and copy data from host to device
   int *dev_in1;
   checkCudaErrors(cudaMalloc((void**)&dev_in1, sizeVec));
   checkCudaErrors(cudaMemcpy(dev_in1, input1, sizeVec, cudaMemcpyHostToDevice));
   
   // memory allocation for dev_in2 and copy data from host to device	
   int *dev_in2;
   checkCudaErrors(cudaMalloc((void**)&dev_in2, sizeVec));
   checkCudaErrors(cudaMemcpy(dev_in2, input2, sizeVec, cudaMemcpyHostToDevice));
   
   // memory allocation for dev_out (no copying as it will be written)
   int *dev_out;
   checkCudaErrors(cudaMalloc((void**)&dev_out, sizeVec));
   
   // kernel call, asynchronous
   MapKernel_CU<<<numBlocks,numThreads>>>(mfunc, dev_in1, dev_in2, dev_out, size);
   
   // block for kernel finish (usually avoid to overlap but here we dont do anything else)
   cudaThreadSynchronize();
   
   // copy back output from device to host memory
   checkCudaErrors(cudaMemcpy(output, dev_out, sizeVec, cudaMemcpyDeviceToHost));
}

int main()
{
   const int SIZE=2<<14;
   int input1[SIZE], input2[SIZE], output[SIZE];
   
   for(int i=0;i<SIZE;i++)
   {
      input1[i]=2;
      input2[i]=5;
      output[i]=0;
   }
   
   MapFunc mapFun;
   map_func(input1, input2, output, SIZE, mapFun);
   
   int total=0;
   for(int i=0;i<SIZE;i++)
   {
      total += output[i];
   }	
   
   std::cout<<"Total: "<<total<<"\n";
   // OUTPUT: Total: 229376 
}